@@ -1,32 +1,45 @@
 package ttlcache
 
 import (
+	"errors"
 	"sync"
 	"time"
 )
 
 // CheckExpireCallback is used as a callback for an external check on item expiration
-type checkExpireCallback func(key string, value interface{}) bool
+type CheckExpireCallback func(key string, value interface{}) bool
 
 // ExpireCallback is used as a callback on item expiration or when notifying of an item new to the cache
-type expireCallback func(key string, value interface{})
+type ExpireCallback func(key string, value interface{})
+
+// LoaderFunction can be supplied to retrieve an item where a cache miss occurs. Supply an item specific ttl or Duration.Zero
+type LoaderFunction func(key string) (data interface{}, ttl time.Duration, err error)
 
 // Cache is a synchronized map of items that can auto-expire once stale
 type Cache struct {
 	mutex                  sync.Mutex
 	ttl                    time.Duration
 	items                  map[string]*item
-	expireCallback         expireCallback
-	checkExpireCallback    checkExpireCallback
-	newItemCallback        expireCallback
+	loaderLock             map[string]*sync.Cond
+	expireCallback         ExpireCallback
+	checkExpireCallback    CheckExpireCallback
+	newItemCallback        ExpireCallback
 	priorityQueue          *priorityQueue
 	expirationNotification chan bool
 	expirationTime         time.Time
 	skipTTLExtension       bool
 	shutdownSignal         chan (chan struct{})
 	isShutDown             bool
+	loaderFunction         LoaderFunction
 }
 
+var (
+	// ErrClosed is raised when operating on a cache where Close() has already been called.
+	ErrClosed = errors.New("cache already closed")
+	// ErrNotFound indicates that the requested key is not present in the cache
+	ErrNotFound = errors.New("key not found")
+)
+
 func (cache *Cache) getItem(key string) (*item, bool, bool) {
 	item, exists := cache.items[key]
 	if !exists || item.expired() {
@@ -105,16 +118,21 @@ func (cache *Cache) startExpirationProcessing() {
 	}
 }
 
+func (cache *Cache) removeItem(item *item) {
+	if cache.expireCallback != nil {
+		go cache.expireCallback(item.key, item.data)
+	}
+	cache.priorityQueue.remove(item)
+	delete(cache.items, item.key)
+
+}
+
 func (cache *Cache) evictjob() {
 	// index will only be advanced if the current entry will not be evicted
 	i := 0
 	for item := cache.priorityQueue.items[i]; ; item = cache.priorityQueue.items[i] {
 
-		cache.priorityQueue.remove(item)
-		delete(cache.items, item.key)
-		if cache.expireCallback != nil {
-			go cache.expireCallback(item.key, item.data)
-		}
+		cache.removeItem(item)
 		if cache.priorityQueue.Len() == 0 {
 			return
 		}
@@ -138,20 +156,16 @@ func (cache *Cache) cleanjob() {
 			}
 		}
 
-		cache.priorityQueue.remove(item)
-		delete(cache.items, item.key)
-		if cache.expireCallback != nil {
-			go cache.expireCallback(item.key, item.data)
-		}
+		cache.removeItem(item)
 		if cache.priorityQueue.Len() == 0 {
 			return
 		}
 	}
 }
 
-// Close calls Purge, and then stops the goroutine that does ttl checking, for a clean shutdown.
-// The cache is no longer cleaning up after the first call to Close, repeated calls are safe though.
-func (cache *Cache) Close() {
+// Close calls Purge after stopping the goroutine that does ttl checking, for a clean shutdown.
+// The cache is no longer cleaning up after the first call to Close, repeated calls are safe and return ErrClosed.
+func (cache *Cache) Close() error {
 
 	cache.mutex.Lock()
 	if !cache.isShutDown {
@@ -163,18 +177,24 @@ func (cache *Cache) Close() {
 		close(cache.shutdownSignal)
 	} else {
 		cache.mutex.Unlock()
+		return ErrClosed
 	}
 	cache.Purge()
+	return nil
 }
 
-// Set is a thread-safe way to add new items to the map
-func (cache *Cache) Set(key string, data interface{}) {
-	cache.SetWithTTL(key, data, ItemExpireWithGlobalTTL)
+// Set is a thread-safe way to add new items to the map.
+func (cache *Cache) Set(key string, data interface{}) error {
+	return cache.SetWithTTL(key, data, ItemExpireWithGlobalTTL)
 }
 
-// SetWithTTL is a thread-safe way to add new items to the map with individual ttl
-func (cache *Cache) SetWithTTL(key string, data interface{}, ttl time.Duration) {
+// SetWithTTL is a thread-safe way to add new items to the map with individual ttl.
+func (cache *Cache) SetWithTTL(key string, data interface{}, ttl time.Duration) error {
 	cache.mutex.Lock()
+	if cache.isShutDown {
+		cache.mutex.Unlock()
+		return ErrClosed
+	}
 	item, exists, _ := cache.getItem(key)
 
 	if exists {
@@ -203,83 +223,164 @@ func (cache *Cache) SetWithTTL(key string, data interface{}, ttl time.Duration)
 		cache.newItemCallback(key, data)
 	}
 	cache.expirationNotification <- true
+	return nil
 }
 
 // Get is a thread-safe way to lookup items
 // Every lookup, also touches the item, hence extending it's life
-func (cache *Cache) Get(key string) (interface{}, bool) {
+func (cache *Cache) Get(key string) (interface{}, error) {
 	cache.mutex.Lock()
+	if cache.isShutDown {
+		cache.mutex.Unlock()
+		return nil, ErrClosed
+	}
 	item, exists, triggerExpirationNotification := cache.getItem(key)
 
 	var dataToReturn interface{}
 	if exists {
 		dataToReturn = item.data
 	}
-	cache.mutex.Unlock()
+
+	var err error = nil
+	if !exists {
+		err = ErrNotFound
+	}
+	if cache.loaderFunction == nil || exists {
+		cache.mutex.Unlock()
+	}
+
+	if cache.loaderFunction != nil && !exists {
+		if lock, ok := cache.loaderLock[key]; ok {
+			// if a lock is present then a fetch is in progress and we wait.
+			cache.mutex.Unlock()
+			lock.L.Lock()
+			lock.Wait()
+			lock.L.Unlock()
+			cache.mutex.Lock()
+			item, exists, triggerExpirationNotification = cache.getItem(key)
+			if exists {
+				dataToReturn = item.data
+				err = nil
+			}
+			cache.mutex.Unlock()
+		} else {
+			// if no lock is present we are the leader and should set the lock and fetch.
+			m := sync.NewCond(&sync.Mutex{})
+			cache.loaderLock[key] = m
+			cache.mutex.Unlock()
+			// cache is not blocked during IO
+			dataToReturn, err = cache.invokeLoader(key)
+			cache.mutex.Lock()
+			m.Broadcast()
+			// cleanup so that we don't block consecutive access.
+			delete(cache.loaderLock, key)
+			cache.mutex.Unlock()
+		}
+
+	}
+
 	if triggerExpirationNotification {
 		cache.expirationNotification <- true
 	}
-	return dataToReturn, exists
+
+	return dataToReturn, err
 }
 
-func (cache *Cache) Remove(key string) bool {
+func (cache *Cache) invokeLoader(key string) (dataToReturn interface{}, err error) {
+	var ttl time.Duration
+
+	dataToReturn, ttl, err = cache.loaderFunction(key)
+	if err == nil {
+		err = cache.SetWithTTL(key, dataToReturn, ttl)
+		if err != nil {
+			dataToReturn = nil
+		}
+	}
+	return dataToReturn, err
+}
+
+// Remove removes an item from the cache if it exists, triggers expiration callback when set. Can return ErrNotFound if the entry was not present.
+func (cache *Cache) Remove(key string) error {
 	cache.mutex.Lock()
+	defer cache.mutex.Unlock()
+	if cache.isShutDown {
+		return ErrClosed
+	}
+
 	object, exists := cache.items[key]
 	if !exists {
-		cache.mutex.Unlock()
-		return false
+		return ErrNotFound
 	}
-	delete(cache.items, object.key)
-	cache.priorityQueue.remove(object)
-	cache.mutex.Unlock()
+	cache.removeItem(object)
 
-	return true
+	return nil
 }
 
-// Count returns the number of items in the cache
+// Count returns the number of items in the cache. Returns zero when the cache has been closed.
 func (cache *Cache) Count() int {
 	cache.mutex.Lock()
+	defer cache.mutex.Unlock()
+
+	if cache.isShutDown {
+		return 0
+	}
 	length := len(cache.items)
-	cache.mutex.Unlock()
 	return length
 }
 
-func (cache *Cache) SetTTL(ttl time.Duration) {
+// SetTTL sets the global TTL value for items in the cache, which can be overridden at the item level.
+func (cache *Cache) SetTTL(ttl time.Duration) error {
 	cache.mutex.Lock()
+
+	if cache.isShutDown {
+		cache.mutex.Unlock()
+		return ErrClosed
+	}
 	cache.ttl = ttl
 	cache.mutex.Unlock()
 	cache.expirationNotification <- true
+	return nil
 }
 
 // SetExpirationCallback sets a callback that will be called when an item expires
-func (cache *Cache) SetExpirationCallback(callback expireCallback) {
+func (cache *Cache) SetExpirationCallback(callback ExpireCallback) {
 	cache.expireCallback = callback
 }
 
 // SetCheckExpirationCallback sets a callback that will be called when an item is about to expire
 // in order to allow external code to decide whether the item expires or remains for another TTL cycle
-func (cache *Cache) SetCheckExpirationCallback(callback checkExpireCallback) {
+func (cache *Cache) SetCheckExpirationCallback(callback CheckExpireCallback) {
 	cache.checkExpireCallback = callback
 }
 
 // SetNewItemCallback sets a callback that will be called when a new item is added to the cache
-func (cache *Cache) SetNewItemCallback(callback expireCallback) {
+func (cache *Cache) SetNewItemCallback(callback ExpireCallback) {
 	cache.newItemCallback = callback
 }
 
-// SkipTtlExtensionOnHit allows the user to change the cache behaviour. When this flag is set to true it will
+// SkipTTLExtensionOnHit allows the user to change the cache behaviour. When this flag is set to true it will
 // no longer extend TTL of items when they are retrieved using Get, or when their expiration condition is evaluated
 // using SetCheckExpirationCallback.
-func (cache *Cache) SkipTtlExtensionOnHit(value bool) {
+func (cache *Cache) SkipTTLExtensionOnHit(value bool) {
 	cache.skipTTLExtension = value
 }
 
+// SetLoaderFunction allows you to set a function to retrieve cache misses. The signature matches that of the Get function.
+// Additional Get calls on the same key block while fetching is in progress (groupcache style).
+func (cache *Cache) SetLoaderFunction(loader LoaderFunction) {
+	cache.loaderFunction = loader
+}
+
 // Purge will remove all entries
-func (cache *Cache) Purge() {
+func (cache *Cache) Purge() error {
 	cache.mutex.Lock()
+	defer cache.mutex.Unlock()
+	if cache.isShutDown {
+		return ErrClosed
+	}
 	cache.items = make(map[string]*item)
 	cache.priorityQueue = newPriorityQueue()
-	cache.mutex.Unlock()
+	return nil
 }
 
 // NewCache is a helper to create instance of the Cache struct
@@ -289,11 +390,13 @@ func NewCache() *Cache {
 
 	cache := &Cache{
 		items:                  make(map[string]*item),
+		loaderLock:             make(map[string]*sync.Cond),
 		priorityQueue:          newPriorityQueue(),
 		expirationNotification: make(chan bool),
 		expirationTime:         time.Now(),
 		shutdownSignal:         shutdownChan,
 		isShutDown:             false,
+		loaderFunction:         nil,
 	}
 	go cache.startExpirationProcessing()
 	return cache
