@@ -1,7 +1,8 @@
-package ttlcache
+package ttlcache_test
 
 import (
 	"math/rand"
+	"sync/atomic"
 	"testing"
 	"time"
 
@@ -10,6 +11,7 @@ import (
 	"fmt"
 	"sync"
 
+	. "github.com/ReneKroon/ttlcache/v2"
 	"github.com/stretchr/testify/assert"
 )
 
@@ -17,9 +19,103 @@ func TestMain(m *testing.M) {
 	goleak.VerifyTestMain(m)
 }
 
+// Issue #31: Test that a single fetch is executed with the loader function
+func TestCache_TestSingleFetch(t *testing.T) {
+	t.Parallel()
+	cache := NewCache()
+	defer cache.Close()
+
+	var calls int32
+
+	loader := func(key string) (data interface{}, ttl time.Duration, err error) {
+		time.Sleep(time.Millisecond * 100)
+		atomic.AddInt32(&calls, 1)
+		return "data", 0, nil
+
+	}
+
+	cache.SetLoaderFunction(loader)
+	wg := sync.WaitGroup{}
+
+	for i := 0; i < 1000; i++ {
+		wg.Add(1)
+		go func() {
+			cache.Get("1")
+			wg.Done()
+		}()
+	}
+	wg.Wait()
+
+	assert.Equal(t, int32(1), calls)
+}
+
+// Issue #30: Removal does not use expiration callback.
+func TestCache_TestRemovalTriggersCallback(t *testing.T) {
+	t.Parallel()
+	cache := NewCache()
+	defer cache.Close()
+
+	var sync = make(chan struct{})
+	expiration := func(key string, data interface{}) {
+
+		sync <- struct{}{}
+	}
+	cache.SetExpirationCallback(expiration)
+
+	cache.Set("1", "barf")
+	cache.Remove("1")
+
+	<-sync
+}
+
+// Issue #31: loader function
+func TestCache_TestLoaderFunction(t *testing.T) {
+	t.Parallel()
+	cache := NewCache()
+
+	cache.SetLoaderFunction(func(key string) (data interface{}, ttl time.Duration, err error) {
+		return nil, 0, ErrNotFound
+	})
+
+	_, err := cache.Get("1")
+	assert.Equal(t, ErrNotFound, err)
+
+	cache.SetLoaderFunction(func(key string) (data interface{}, ttl time.Duration, err error) {
+		return "1", 0, nil
+	})
+
+	value, found := cache.Get("1")
+	assert.Equal(t, nil, found)
+	assert.Equal(t, "1", value)
+
+	cache.Close()
+
+	value, found = cache.Get("1")
+	assert.Equal(t, ErrClosed, found)
+	assert.Equal(t, nil, value)
+}
+
+// Issue #31: edge case where cache is closed when loader function has completed
+func TestCache_TestLoaderFunctionDuringClose(t *testing.T) {
+	t.Parallel()
+	cache := NewCache()
+
+	cache.SetLoaderFunction(func(key string) (data interface{}, ttl time.Duration, err error) {
+		cache.Close()
+		return "1", 0, nil
+	})
+
+	value, found := cache.Get("1")
+	assert.Equal(t, ErrClosed, found)
+	assert.Equal(t, nil, value)
+
+	cache.Close()
+
+}
+
 // Issue #28: call expirationCallback automatically on cache.Close()
 func TestCache_ExpirationOnClose(t *testing.T) {
-
+	t.Parallel()
 	cache := NewCache()
 
 	success := make(chan struct{})
@@ -49,8 +145,9 @@ func TestCache_ExpirationOnClose(t *testing.T) {
 }
 
 // # Issue 29: After Close() the behaviour of Get, Set, Remove is not defined.
-/*
+
 func TestCache_ModifyAfterClose(t *testing.T) {
+	t.Parallel()
 	cache := NewCache()
 
 	cache.SetTTL(time.Hour * 100)
@@ -61,75 +158,84 @@ func TestCache_ModifyAfterClose(t *testing.T) {
 	cache.Set("2", 1)
 	cache.Set("3", 1)
 
+	_, findErr := cache.Get("1")
+	assert.Equal(t, nil, findErr)
+	assert.Equal(t, nil, cache.Set("broken", 1))
+	assert.Equal(t, ErrNotFound, cache.Remove("broken2"))
+	assert.Equal(t, nil, cache.Purge())
+	assert.Equal(t, nil, cache.SetWithTTL("broken", 2, time.Minute))
+	assert.Equal(t, nil, cache.SetTTL(time.Hour))
+
 	cache.Close()
 
-	cache.Get("broken3")
-	cache.Set("broken", 1)
-	cache.Remove("broken2")
+	_, getErr := cache.Get("broken3")
+	assert.Equal(t, ErrClosed, getErr)
+	assert.Equal(t, ErrClosed, cache.Set("broken", 1))
+	assert.Equal(t, ErrClosed, cache.Remove("broken2"))
+	assert.Equal(t, ErrClosed, cache.Purge())
+	assert.Equal(t, ErrClosed, cache.SetWithTTL("broken", 2, time.Minute))
+	assert.Equal(t, ErrClosed, cache.SetTTL(time.Hour))
+	assert.Equal(t, 0, cache.Count())
 
-	wait := time.NewTimer(time.Millisecond * 100)
-
-	select {
-	case <-wait.C:
-		t.Fail()
-	}
-
-}*/
+}
 
 // Issue #23: Goroutine leak on closing. When adding a close method i would like to see
 // that it can be called in a repeated way without problems.
 func TestCache_MultipleCloseCalls(t *testing.T) {
+	t.Parallel()
 	cache := NewCache()
 
 	cache.SetTTL(time.Millisecond * 100)
 
-	cache.SkipTtlExtensionOnHit(false)
+	cache.SkipTTLExtensionOnHit(false)
 	cache.Set("test", "!")
 	startTime := time.Now()
 	for now := time.Now(); now.Before(startTime.Add(time.Second * 3)); now = time.Now() {
-		if _, found := cache.Get("test"); !found {
+		if _, err := cache.Get("test"); err != nil {
 			t.Errorf("Item was not found, even though it should not expire.")
 		}
 
 	}
 
 	cache.Close()
-	cache.Close()
-	cache.Close()
-	cache.Close()
+	assert.Equal(t, ErrClosed, cache.Close())
 }
 
 // test for Feature request in issue #12
 //
 func TestCache_SkipTtlExtensionOnHit(t *testing.T) {
+	t.Parallel()
+
 	cache := NewCache()
 	defer cache.Close()
 
 	cache.SetTTL(time.Millisecond * 100)
 
-	cache.SkipTtlExtensionOnHit(false)
+	cache.SkipTTLExtensionOnHit(false)
 	cache.Set("test", "!")
 	startTime := time.Now()
 	for now := time.Now(); now.Before(startTime.Add(time.Second * 3)); now = time.Now() {
-		if _, found := cache.Get("test"); !found {
+		if _, err := cache.Get("test"); err != nil {
 			t.Errorf("Item was not found, even though it should not expire.")
 		}
 
 	}
 
-	cache.SkipTtlExtensionOnHit(true)
+	cache.SkipTTLExtensionOnHit(true)
 	cache.Set("expireTest", "!")
 	// will loop if item does not expire
-	for _, found := cache.Get("expireTest"); found; _, found = cache.Get("expireTest") {
+	for _, err := cache.Get("expireTest"); err == nil; _, err = cache.Get("expireTest") {
 	}
 }
 
 func TestCache_ForRacesAcrossGoroutines(t *testing.T) {
+	t.Parallel()
+
 	cache := NewCache()
 	defer cache.Close()
 
 	cache.SetTTL(time.Minute * 1)
-	cache.SkipTtlExtensionOnHit(false)
+	cache.SkipTTLExtensionOnHit(false)
 
 	var wgSet sync.WaitGroup
 	var wgGet sync.WaitGroup
@@ -175,7 +281,7 @@ func TestCache_SkipTtlExtensionOnHit_ForRacesAcrossGoroutines(t *testing.T) {
 	defer cache.Close()
 
 	cache.SetTTL(time.Minute * 1)
-	cache.SkipTtlExtensionOnHit(true)
+	cache.SkipTTLExtensionOnHit(true)
 
 	var wgSet sync.WaitGroup
 	var wgGet sync.WaitGroup
@@ -219,6 +325,8 @@ func TestCache_SkipTtlExtensionOnHit_ForRacesAcrossGoroutines(t *testing.T) {
 // test github issue #14
 // Testing expiration callback would continue with the next item in list, even when it exceeds list lengths
 func TestCache_SetCheckExpirationCallback(t *testing.T) {
+	t.Parallel()
+
 	iterated := 0
 	ch := make(chan struct{})
 
@@ -250,6 +358,7 @@ func TestCache_SetCheckExpirationCallback(t *testing.T) {
 // Which is not right when we become negative due to scheduling.
 // This test could use improvement as it's not requiring a lot of time to trigger.
 func TestCache_SetExpirationCallback(t *testing.T) {
+	t.Parallel()
 
 	type A struct {
 	}
@@ -275,6 +384,8 @@ func TestCache_SetExpirationCallback(t *testing.T) {
 
 // test github issue #4
 func TestRemovalAndCountDoesNotPanic(t *testing.T) {
+	t.Parallel()
+
 	cache := NewCache()
 	defer cache.Close()
 
@@ -286,6 +397,8 @@ func TestRemovalAndCountDoesNotPanic(t *testing.T) {
 
 // test github issue #3
 func TestRemovalWithTtlDoesNotPanic(t *testing.T) {
+	t.Parallel()
+
 	cache := NewCache()
 	defer cache.Close()
 
@@ -297,8 +410,8 @@ func TestRemovalWithTtlDoesNotPanic(t *testing.T) {
 	cache.Set("key", "value")
 	cache.Remove("key")
 
-	value, exists := cache.Get("keyWithTTL")
-	if exists {
+	value, err := cache.Get("keyWithTTL")
+	if err == nil {
 		t.Logf("got %s for keyWithTTL\n", value)
 	}
 	count := cache.Count()
@@ -306,8 +419,8 @@ func TestRemovalWithTtlDoesNotPanic(t *testing.T) {
 
 	<-time.After(3 * time.Second)
 
-	value, exists = cache.Get("keyWithTTL")
-	if exists {
+	value, err = cache.Get("keyWithTTL")
+	if err != nil {
 		t.Logf("got %s for keyWithTTL\n", value)
 	} else {
 		t.Logf("keyWithTTL has gone")
@@ -317,6 +430,8 @@ func TestRemovalWithTtlDoesNotPanic(t *testing.T) {
 }
 
 func TestCacheIndividualExpirationBiggerThanGlobal(t *testing.T) {
+	t.Parallel()
+
 	cache := NewCache()
 	defer cache.Close()
 
@@ -324,32 +439,36 @@ func TestCacheIndividualExpirationBiggerThanGlobal(t *testing.T) {
 	cache.SetWithTTL("key", "value", time.Duration(100*time.Millisecond))
 	<-time.After(150 * time.Millisecond)
 	data, exists := cache.Get("key")
-	assert.Equal(t, exists, false, "Expected item to not exist")
+	assert.Equal(t, exists, ErrNotFound, "Expected item to not exist")
 	assert.Nil(t, data, "Expected item to be nil")
 }
 
 func TestCacheGlobalExpirationByGlobal(t *testing.T) {
+	t.Parallel()
+
 	cache := NewCache()
 	defer cache.Close()
 
 	cache.Set("key", "value")
 	<-time.After(50 * time.Millisecond)
 	data, exists := cache.Get("key")
-	assert.Equal(t, exists, true, "Expected item to exist in cache")
+	assert.Equal(t, exists, nil, "Expected item to exist in cache")
 	assert.Equal(t, data.(string), "value", "Expected item to have 'value' in value")
 
 	cache.SetTTL(time.Duration(50 * time.Millisecond))
 	data, exists = cache.Get("key")
-	assert.Equal(t, exists, true, "Expected item to exist in cache")
+	assert.Equal(t, exists, nil, "Expected item to exist in cache")
 	assert.Equal(t, data.(string), "value", "Expected item to have 'value' in value")
 
 	<-time.After(100 * time.Millisecond)
 	data, exists = cache.Get("key")
-	assert.Equal(t, exists, false, "Expected item to not exist")
+	assert.Equal(t, exists, ErrNotFound, "Expected item to not exist")
 	assert.Nil(t, data, "Expected item to be nil")
 }
 
 func TestCacheGlobalExpiration(t *testing.T) {
+	t.Parallel()
+
 	cache := NewCache()
 	defer cache.Close()
 
@@ -358,10 +477,12 @@ func TestCacheGlobalExpiration(t *testing.T) {
 	cache.Set("key_2", "value")
 	<-time.After(200 * time.Millisecond)
 	assert.Equal(t, 0, cache.Count(), "Cache should be empty")
-	assert.Equal(t, 0, cache.priorityQueue.Len(), "PriorityQueue should be empty")
+
 }
 
 func TestCacheMixedExpirations(t *testing.T) {
+	t.Parallel()
+
 	cache := NewCache()
 	defer cache.Close()
 
@@ -376,6 +497,8 @@ func TestCacheMixedExpirations(t *testing.T) {
 }
 
 func TestCacheIndividualExpiration(t *testing.T) {
+	t.Parallel()
+
 	cache := NewCache()
 	defer cache.Close()
 
@@ -394,21 +517,25 @@ func TestCacheIndividualExpiration(t *testing.T) {
 }
 
 func TestCacheGet(t *testing.T) {
+	t.Parallel()
+
 	cache := NewCache()
 	defer cache.Close()
 
 	data, exists := cache.Get("hello")
-	assert.Equal(t, exists, false, "Expected empty cache to return no data")
+	assert.Equal(t, exists, ErrNotFound, "Expected empty cache to return no data")
 	assert.Nil(t, data, "Expected data to be empty")
 
 	cache.Set("hello", "world")
 	data, exists = cache.Get("hello")
 	assert.NotNil(t, data, "Expected data to be not nil")
-	assert.Equal(t, true, exists, "Expected data to exist")
+	assert.Equal(t, nil, exists, "Expected data to exist")
 	assert.Equal(t, "world", (data.(string)), "Expected data content to be 'world'")
 }
 
 func TestCacheExpirationCallbackFunction(t *testing.T) {
+	t.Parallel()
+
 	expiredCount := 0
 	var lock sync.Mutex
 
@@ -433,13 +560,15 @@ func TestCacheExpirationCallbackFunction(t *testing.T) {
 // TestCacheCheckExpirationCallbackFunction should consider that the next entry in the queue
 // needs to be considered for eviction even if the callback returns no eviction for the current item
 func TestCacheCheckExpirationCallbackFunction(t *testing.T) {
+	t.Parallel()
+
 	expiredCount := 0
 	var lock sync.Mutex
 
 	cache := NewCache()
 	defer cache.Close()
 
-	cache.SkipTtlExtensionOnHit(true)
+	cache.SkipTTLExtensionOnHit(true)
 	cache.SetTTL(time.Duration(50 * time.Millisecond))
 	cache.SetCheckExpirationCallback(func(key string, value interface{}) bool {
 		if key == "key2" || key == "key4" {
@@ -464,6 +593,8 @@ func TestCacheCheckExpirationCallbackFunction(t *testing.T) {
 }
 
 func TestCacheNewItemCallbackFunction(t *testing.T) {
+	t.Parallel()
+
 	newItemCount := 0
 	cache := NewCache()
 	defer cache.Close()
@@ -480,6 +611,8 @@ func TestCacheNewItemCallbackFunction(t *testing.T) {
 }
 
 func TestCacheRemove(t *testing.T) {
+	t.Parallel()
+
 	cache := NewCache()
 	defer cache.Close()
 
@@ -489,11 +622,13 @@ func TestCacheRemove(t *testing.T) {
 	<-time.After(70 * time.Millisecond)
 	removeKey := cache.Remove("key")
 	removeKey2 := cache.Remove("key_2")
-	assert.Equal(t, true, removeKey, "Expected 'key' to be removed from cache")
-	assert.Equal(t, false, removeKey2, "Expected 'key_2' to already be expired from cache")
+	assert.Equal(t, nil, removeKey, "Expected 'key' to be removed from cache")
+	assert.Equal(t, ErrNotFound, removeKey2, "Expected 'key_2' to already be expired from cache")
 }
 
 func TestCacheSetWithTTLExistItem(t *testing.T) {
+	t.Parallel()
+
 	cache := NewCache()
 	defer cache.Close()
 
@@ -502,11 +637,13 @@ func TestCacheSetWithTTLExistItem(t *testing.T) {
 	<-time.After(30 * time.Millisecond)
 	cache.SetWithTTL("key", "value2", time.Duration(50*time.Millisecond))
 	data, exists := cache.Get("key")
-	assert.Equal(t, true, exists, "Expected 'key' to exist")
+	assert.Equal(t, nil, exists, "Expected 'key' to exist")
 	assert.Equal(t, "value2", data.(string), "Expected 'data' to have value 'value2'")
 }
 
 func TestCache_Purge(t *testing.T) {
+	t.Parallel()
+
 	cache := NewCache()
 	defer cache.Close()
 
